/* Copyright (c) Determina Corp. 2006.  All rights reserved.
 * 
 * This software and documentation constitute an unpublished work and contain
 * valuable trade secrets and proprietary information belonging to Determina
 * Corp.  None of the foregoing material may be copied, duplicated or
 * disclosed without the express written permission of Determina Corp.
 * DETERMINA CORP. EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING THIS
 * SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF MERCHANTABILITY
 * AND/OR FITNESS FOR ANY PARTICULAR PURPOSE, AND WARRANTIES OF PERFORMANCE,
 * AND ANY WARRANTY THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE
 * OF TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO THE USE
 * OF THE SOFTWARE OR DOCUMENTATION. Under no circumstances shall Determina
 * Corp. be liable for incidental, special, indirect, direct or consequential
 * damages or loss of profits, interruption of business, or related expenses
 * which may arise from use of software or documentation, including but not
 * limited to those resulting from defects in software and/or documentation,
 * or loss or inaccuracy of data of any kind.
 *
 * CVE-2006-3730 fix
 *
 */


#include <windows.h>
#include <winnt.h>

//
// Structures for module matching
//

struct hash {
    unsigned int start;
    unsigned int length;
    unsigned int crc32;
};

struct module {
    char* name;                 // DLL name
    struct hash hash;           // CRC32 checksum of the area around the patch
    unsigned int patch_point;   // offset of the patch point
};


//
// Modules that we watch to patch
//

struct module modules[] = {

    // Windows 2000 SP0

    {
        "webvw.DLL",
        { 0xDA6B, 0x22, 3206657593 },
        0xda8b
    },

    // Windows 2000 MS04-037

    {
        "webvw.DLL",
        { 0x70A9, 0x22, 1929135436 },
        0x70c9
    },

    // Windows 2000 SP4 UR1

    {
         "webvw.DLL",
        { 0x70A9, 0x22, 1929135436 },
        0x70c9
    },

    // Windows 2000 MS05-024

    {
        "webvw.DLL",
        { 0x70D9, 0x22, 2633833585 },
        0x70f9
    },

    // Windows 2000 MS05-049

    {
        "webvw.DLL",
        { 0x7129, 0x22, 1929135436 },
        0x7149
    },

    // Windows XP SP0

    {
        "webvw.dll",
        { 0x53FD, 0x22, 2807024744 },
        0x541d
    },

    // Windows XP SP1a

    {
        "webvw.dll",
        { 0x5413, 0x22, 2671017242 },
        0x5433
    },

    // Windows XP SP2

    {
        "webvw.dll",
        { 0x89E3, 0x40, 2830078522 },
        0x8A21
    },

    // Windows 2003 SP0

    { "webvw.dll",
       { 0x589A, 0x22, 2033975504 },
      0x58ba
    },

    // Windows 2003 SP1

    { "webvw.dll",
       { 0x8D52, 0x22, 1974643245 },
      0x8d72
    }
};

#define NUM_MODULES (sizeof(modules)/sizeof(struct module))

// Patch status

enum patch_status {
    NOT_PATCHED,
    PATCHED
};

extern "C" {
    __declspec(dllexport) enum patch_status patch_status = NOT_PATCHED;
}


// Address of the DSA_SetItem function in WEBVW.DLL
void* DSA_SetItem = NULL;

//
// DSA_SetItem hook
//
// This hook is intercepts calls to the DSA_SetItem function, documented here:
// http://windowssdk.msdn.microsoft.com/en-us/library/ms672647.aspx
//
// The ReactOS source code of the function is available here:
// http://www.reactos.org/generated/doxygen/d9/df2/dsa_8c-source.html
//
// The index parameter is at ESP+8. We need to ensure that (index+2) * 16 does
// not overflow. The smallest index that triggers the overflow is 0xFFFFFFE.
// If the index parameter is not less than that value, we'll set it to
// 0x80000000, which is a negative number and will cause DSA_SetItem to abort.

__declspec(naked) void DSA_SetItem_hook()
{
    __asm {
        push eax                // save eax
        mov eax, 0xffffffe      // minimum size that causes the overflow
        cmp [esp+0xC], eax      // if (index < 0xffffffe)
        jb hook_return          //     goto hook_return;
        mov eax, 0x80000000     // a negative index will cause DSA_SetItem to abort
        mov [esp+0xC], eax      // index = 0x80000000;
hook_return:
        pop eax                 // restore eax
        jmp [DSA_SetItem]       // jump to the DSA_SetItem function
    }
}

// Address of the DSA_SetItem hook function
void* pDSA_SetItem_hook = &DSA_SetItem_hook;

// We're going to patch the CWebViewFolderIcon::setSlice function in WEBVW.DLL.
//
// 5a6a8a21 ff15508a6a5a     call dword ptr [webvw!_imp__DSA_SetItem (5a6a8a50)]
//
// Our patch will modify the operand of the call instruction and replace it
// with the address of the pDSA_SetItem_hook variable. The patched call
// instruction will dereference that variable and jump to the DSA_SetItem_hook
// function.
//
// DSA_SetItem_hook will fix the index argument and jump to the original
// destination of the call.

// Opcode of the instruction we're patching
#define CALL_DS "\xFF\x15"      // call ds:address

//
// Patch the module. Returns FALSE if the patch could not be applied.
//

BOOL patch_module(struct module* module, void* image_base)
{
    // Get the absolute addresses of the patch point

    char* patch_point = (char*)image_base + module->patch_point;

    // Make sure that the patch point starts with a call ds:address instruction

    if (memcmp(patch_point, CALL_DS, sizeof(CALL_DS)-1) != 0)
        return FALSE;
    
    // We want to patch the address in the call instruction

    patch_point += 2;
    
    // Get the address of the DSA_SetItem function

    DSA_SetItem = **(void***)patch_point;
    
    // Change the memory protection of the patch point to enable writing
    
    DWORD old_protection, dummy;

    if (VirtualProtect(patch_point, sizeof(void*), PAGE_EXECUTE_READWRITE, &old_protection) == 0) {
        return FALSE;
    }

    // Modify the call instruction to use pDSA_SetItem_hook
    *(void**)patch_point = &pDSA_SetItem_hook;

    // Restore the memory protection of the patch point

    VirtualProtect(patch_point, sizeof(void*), old_protection, &dummy);

    patch_status = PATCHED;

    return TRUE;
}


//
// Table for CRC32 calculation
//

const unsigned int crctab[] = {
    0x00000000, 0x77073096, 0xee0e612c, 0x990951ba,
    0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
    0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
    0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
    0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
    0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
    0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
    0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
    0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
    0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
    0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
    0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
    0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
    0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
    0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
    0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
    0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
    0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
    0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
    0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
    0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
    0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
    0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
    0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
    0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
    0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
    0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
    0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
    0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
    0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
    0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
    0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
    0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
    0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
    0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
    0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
    0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
    0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
    0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
    0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
    0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
    0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
    0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
    0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
    0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
    0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
    0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
    0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
    0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
    0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
    0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
    0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
    0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
    0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
    0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
    0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
    0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
    0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
    0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
    0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
    0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
    0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
    0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
    0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};


//
// This function implements the Ethernet AUTODIN II CRC32 algorithm.
//

unsigned int crc32(const char* buf, const unsigned int len)
{
    unsigned int i;
    unsigned int crc = 0xFFFFFFFF;

    for (i = 0; i < len; i++)
        crc = (crc >> 8) ^ crctab[(crc ^ buf[i]) & 0xFF];

    return crc;
}


//
// Test if an address is inside a loaded PE image. Returns TRUE or FALSE
//

BOOL addr_in_image(void* addr, void* image_base, DWORD image_size)
{
    if (addr == NULL || image_base == NULL || image_size <= 0)
        return FALSE;

    if ((char*) addr < (char*) image_base ||
        (char*) addr > ((char*) image_base + image_size)) {
        return FALSE;
    }

    // FIXME: do a QueryVirtualMemory() & check for image; to be safe
    return TRUE;
}


//
// Get the DLL name of a loaded PE image. Returns NULL on error
//

char* get_dll_name(void* image_base, DWORD image_size, IMAGE_NT_HEADERS* nt)
{
    IMAGE_DATA_DIRECTORY *expdir = nt->OptionalHeader.DataDirectory +
                                   IMAGE_DIRECTORY_ENTRY_EXPORT;
    IMAGE_EXPORT_DIRECTORY *exports = NULL;
    char* dll_name = NULL;

    if (!addr_in_image((void*) expdir, image_base, image_size))
        return NULL;

    // We have a valid exports directory entry in data direcotry.
    if (expdir->Size > 0) {
        exports = (IMAGE_EXPORT_DIRECTORY *) ((char*) image_base +
                                              expdir->VirtualAddress);
    } else {
        return NULL;
    }

    if (!addr_in_image((void*) exports, image_base, image_size))
        return NULL;

    // We have a valid exports direcotry.
    dll_name =  (char*) image_base + exports->Name /* RVA */;
    if (!addr_in_image((void*) dll_name, image_base, image_size)) {
        return NULL;
    } else {
        return dll_name;
    }
}


//
// Matches the DLL loaded at the specified image base against the list of
// modules we want to patch. If the DLL matches, returns a pointer to the
// module structure, otherwise returns NULL.
//

struct module* module_match(void* image_base)
{
    IMAGE_DOS_HEADER* dos = (IMAGE_DOS_HEADER*) image_base;
    IMAGE_NT_HEADERS* nt;
    DWORD image_size;
    char* dll_name = NULL;
    unsigned int image_crc32 = 0;
    int i;

    // FIXME: assumes readability of header (dos, nt & optional) memory - 
    //      validate.

    // Is this an image of a PE file?
    if (image_base == NULL || dos->e_magic != IMAGE_DOS_SIGNATURE)
        return NULL;

    nt = (IMAGE_NT_HEADERS*) (((unsigned int)dos) + dos->e_lfanew);
    if (nt == NULL || nt->Signature != IMAGE_NT_SIGNATURE)
        return NULL;

    image_size = nt->OptionalHeader.SizeOfImage;
    if (image_size <= 0)
        return NULL;

    // Well, looks like an image, let us see if it has a PE name.
    dll_name = get_dll_name(image_base, image_size, nt);
    if (dll_name == NULL)
        return NULL;

    // Let us see if the image matches any of our DLLs.
    for (i = 0; i < NUM_MODULES; i++) {

        // It is ok to use strlen on the argument because it is a compile time
        // constant; the compiler should replace the call with a constant.
        if (memcmp((void*) dll_name, (void*) modules[i].name,
                   strlen(modules[i].name)) == 0) {
            char *patch_point = (char*)image_base + modules[i].patch_point;
            char *hash_start = (char*)image_base + modules[i].hash.start;
            unsigned int hash_size = modules[i].hash.length;

            // Ok, dll name matched.  Now, let us verify hash bounds.
            if (!addr_in_image((void*) hash_start, image_base, image_size))
                continue;
            if (!addr_in_image((void*) (hash_start + hash_size),
                               image_base, image_size))
                continue;

            // Check the CRC32 hash
            image_crc32 = crc32(hash_start, hash_size);
            if (image_crc32 != modules[i].hash.crc32)
                continue;

            // Verify that the patch point is within bounds
            if (!addr_in_image((void*) (patch_point+10),
                               image_base, image_size))
                continue;

            // We have a matching module
            return &modules[i];
        }
    }

    return NULL;
}


//
// Internal structures used for walking the loaded modules list
//

typedef struct {
    ULONG Length;
    BOOLEAN Initialized;
    PVOID SsHandle;
    LIST_ENTRY InLoadOrderModuleList;
    LIST_ENTRY InMemoryOrderModuleList;
    LIST_ENTRY InInitializationOrderModuleList;
} PEB_LDR_DATA;

typedef struct {
    LIST_ENTRY InLoadOrderModuleList;
    LIST_ENTRY InMemoryOrderModuleList;
    LIST_ENTRY InInitializationOrderModuleList;
    PVOID BaseAddress;
    // incomplete, but we don't need the rest
} LDR_MODULE;


//
// Walk the list of loaded DLLs and patch the ones that match
//

void patch_loaded_modules()
{
    PEB_LDR_DATA *ldr;
    LIST_ENTRY *first, *e;
    LDR_MODULE *mod;

    __asm {
        mov eax, fs:[0x30]  // get the PEB pointer
        mov eax, [eax+0xC]  // get the pointer to the loader data structure
        mov ldr, eax
    }

    first = &ldr->InLoadOrderModuleList;

    int count = 0;

    for (e = first->Flink; e != first; e = e->Flink)
    {
        mod = (LDR_MODULE*)e;

        // Patch the module if it matches

        struct module* module = module_match(mod->BaseAddress);

        if (module != NULL)
            patch_module(module, mod->BaseAddress);

        // Try to detect infinite loops
        
        if (count++ > 10000) {
            return;
        }
    }
}


//
// If our loader hook was successfully inserted, this function will be called
// every time the loader calls the DllMain routine of any loaded DLL. This
// allows us to intercept DLL loading and unloading. The arguments are the same
// as the ones passed to the real DllMain function.
//

void WINAPI DllMainHook(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
{
    UNREFERENCED_PARAMETER(fdwReason);
    UNREFERENCED_PARAMETER(lpvReserved);

    switch (fdwReason) {
        case DLL_PROCESS_ATTACH:

            // If the DLL matches our list of modules, apply the hotpatch
            
            struct module* module = module_match(hinstDll);
            if (module != NULL)
                patch_module(module, hinstDll);
            break;
    }

    return;
}


/* To intercept DLL loading and unloading, we will hook the ntdll function that
   calls the DllMain functions for all DLLs. The body of this function is
   exactly the same on Windows 2000, XP and 2003.

.text:7C901193             _LdrpCallInitRoutine@16 proc near
.text:7C901193
.text:7C901193             arg_0           = dword ptr  8
.text:7C901193             arg_4           = dword ptr  0Ch
.text:7C901193             arg_8           = dword ptr  10h
.text:7C901193             arg_C           = dword ptr  14h
.text:7C901193
.text:7C901193 55                          push    ebp
.text:7C901194 8B EC                       mov     ebp, esp
.text:7C901196 56                          push    esi
.text:7C901197 57                          push    edi
.text:7C901198 53                          push    ebx
.text:7C901199 8B F4                       mov     esi, esp
.text:7C90119B FF 75 14                    push    [ebp+arg_C]
.text:7C90119E FF 75 10                    push    [ebp+arg_8]
.text:7C9011A1 FF 75 0C                    push    [ebp+arg_4]
.text:7C9011A4 FF 55 08                    call    [ebp+arg_0]
.text:7C9011A7 8B E6                       mov     esp, esi
.text:7C9011A9 5B                          pop     ebx
.text:7C9011AA 5F                          pop     edi
.text:7C9011AB 5E                          pop     esi
.text:7C9011AC 5D                          pop     ebp
.text:7C9011AD C2 10 00                    retn    10h
.text:7C9011AD             _LdrpCallInitRoutine@16 endp

   This function is not exported, but there is a simple way to get its address.
   When our own DllMain function is called, the return address of the stack
   will point to the mov esp, esi instruction in LdrpCallInitRoutine. We will
   verify that the 12 bytes before the return address match, and then we will
   overwrite the first two push instructions with a jump to our loader hook
   function. The loader hook function will push the three DllMain arguments,
   call DllMainHook, push the three arguments again and jump back to the
   call [ebp+arg_0] instrcution in ntdll.dll. */


// The 12 bytes at the patch point are three push instructions and a call

const char loader_patch_bytes[] = "\xFF\x75\x14"    // push [ebp+arg_C]
                                  "\xFF\x75\x10"    // push [ebp+arg_8]
                                  "\xFF\x75\x0C"    // push [ebp+arg_4]
                                  "\xFF\x55\x08";   // call [ebp+arg_0]

const int  loader_patch_size = sizeof(loader_patch_bytes)-1;

// Patch point address, NULL if the loader hook is not currently inserted

char* loader_patch_point = NULL;

// This is the address of the call [ebp+arg_0] instrcution in ntdll.dll. The
// loader hook function reads this variable to find out where to return to.

char* loader_hook_return;


//
// Loader hook function, called from LdrpCallInitRoutine
//

__declspec(naked) void loader_hook()
{
    __asm {
        push [ebp+0x14]             // push [ebp+arg_C]
        push [ebp+0x10]             // push [ebp+arg_8]
        push [ebp+0xC]              // push [ebp+arg_4]
        call DllMainHook            // call our DllMain hook
        
        push [ebp+0x14]             // push [ebp+arg_C]
        push [ebp+0x10]             // push [ebp+arg_8]
        push [ebp+0xC]              // push [ebp+arg_4]
        jmp  [loader_hook_return]   // jump to the call [ebp+arg_0] instruction
    }
}

// Relative jump opcode

#define JMP_REL32 '\xE9'

//
// Patch the ntdll!LdrpCallInitRoutine function. The argument is the return
// address from DllMain. Returns FALSE if the patch could not be applied.
//

BOOL insert_loader_hook(char* return_address)
{
    char* patch_point = return_address - loader_patch_size;

    if (memcmp(patch_point, loader_patch_bytes, loader_patch_size) != 0) {
        return FALSE;
    }

    // Store the address of the call [ebp+arg_0] instruction

    loader_hook_return = return_address - 3;

    // Change the memory protection of the patch point to enable writing
    
    DWORD old_protection, dummy;

    if (VirtualProtect(patch_point, 5, PAGE_EXECUTE_READWRITE, &old_protection) == 0) {
        return FALSE;
    }

    // Write a 5 byte relative jump to the loader hook

    int loader_hook_rel = (int)&loader_hook - ((int)patch_point + 5);

            patch_point[0] = JMP_REL32;         // relative jump opcode
    *(int*)&patch_point[1] = loader_hook_rel;   // relative jump target

    loader_patch_point = patch_point;
    
    // Restore the memory protection of the patch point

    VirtualProtect(patch_point, 5, old_protection, &dummy);

    return TRUE;
}


//
// Remove the loader hook, returns FALSE on error
//

BOOL remove_loader_hook()
{
    if (loader_patch_point != NULL) {

        // Change the memory protection of the patch point to enable writing
        
        DWORD old_protection, dummy;

        if (VirtualProtect(loader_patch_point, 5, PAGE_EXECUTE_READWRITE, &old_protection) == 0) {
            return FALSE;
        }

        // Restore the 5 bytes that we overwrote

        memcpy(loader_patch_point, loader_patch_bytes, 5);

        // Restore the memory protection of the patch point

        VirtualProtect(loader_patch_point, 5, old_protection, &dummy);
    }

    return TRUE;
}


//
// DLL entry point
//

BOOL WINAPI DllMain(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
{
    UNREFERENCED_PARAMETER(hinstDll);
    UNREFERENCED_PARAMETER(lpvReserved);

    // Get the return address of the current function
    char* return_address = ((char**)&hinstDll)[-1];

    switch (fdwReason) {
        case DLL_PROCESS_ATTACH:

            // patch the DLLs that have already been loaded
            patch_loaded_modules();

            // insert the loader hook
            insert_loader_hook(return_address);
            break;

        case DLL_PROCESS_DETACH:
            // remove the loader hook
            remove_loader_hook();
            break;
    }

    return TRUE;
}
